# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os

from kivy.app import App
from kivy.logger import Logger
from kivy.clock import Clock
from kivy.properties import (AliasProperty, BooleanProperty, ObjectProperty,
								NumericProperty)
from kivy.resources import resource_find
from .app import ELiDEApp
from kivy.uix.widget import Widget
from kivy.uix.screenmanager import ScreenManager, Screen
import LiSE.proxy
from .util import trigger
from functools import partial


class GameScreen(Screen):
	"""A version of :class:`kivy.uix.screenmanager.Screen` that is easier to set up and use with ELiDE

    Should be a child of the :class:`ELiDE.game.Screens` widget, which will never itself be displayed.
    ``GameScreen`` instances in it will be added to the screen manager, so that you can switch
    to them with the ``switch_screen`` method.

    Every ``GameScreen`` needs a ``name``, just like regular ``Screen``.

    """
	switch_screen = ObjectProperty()
	"""Method to set the ``screen`` attribute of the main :class:`kivy.uix.screenmanager.ScreenManager`"""
	disabled = BooleanProperty(False)
	"""If you bind your widgets' ``disabled`` to this, they will be disabled when a game command is in mid-execution"""

	@property
	def app(self):
		return App.get_running_app()

	@property
	def engine(self):
		return App.get_running_app().engine

	def disable_input(self, cb=None):
		"""Set ``self.disabled`` to ``True``, then call ``cb`` if provided

        :param cb: callback function for after disabling
        :return: ``None``

        """
		self.disabled = True
		if cb:
			cb()

	def enable_input(self, cb=None):
		"""Call ``cb`` if provided, then set ``self.disabled`` to ``False``

        :param cb: callback function for before enabling
        :return: ``None``

        """
		if cb:
			cb()
		self.disabled = False

	def wait_travel(self, character, thing, dest, cb=None):
		"""Schedule a thing to travel someplace, then wait for it to finish.

        :param character: name of the character
        :param thing: name of the thing that will travel
        :param dest: name of the place it will travel to
        :param cb: callback function for when it's done, optional
        :return: ``None``

        """
		self.disable_input()
		self.app.wait_travel(character,
								thing,
								dest,
								cb=partial(self.enable_input, cb))

	def wait_turns(self, turns, cb=None):
		"""Call ``self.app.engine.next_turn()`` ``n`` times, waiting ``self.app.turn_length`` in between

        Disables input for the duration.

        :param turns: number of turns to wait
        :param cb: function to call when done waiting, optional
        :return: ``None``

        """
		self.disable_input()
		self.app.wait_turns(turns, cb=partial(self.enable_input, cb))

	def wait_command(self, start_func, turns=1, end_func=None):
		"""Call ``start_func``, wait ``turns``, and then call ``end_func`` if provided

        Disables input for the duration.

        :param start_func: function to call just after disabling input
        :param turns: number of turns to wait
        :param end_func: function to call just before re-enabling input
        :return: ``None``

        """
		self.disable_input()
		start_func()
		self.app.wait_turns(turns, cb=partial(self.enable_input, end_func))

	def wait_travel_command(self,
							character,
							thing,
							dest,
							start_func,
							turns=1,
							end_func=lambda: None):
		"""Schedule a thing to travel someplace and do something, then wait for it to finish.

        Input will be disabled for the duration.

        :param character: name of the character
        :param thing: name of the thing
        :param dest: name of the destination (a place)
        :param start_func: function to call when the thing gets to dest
        :param turns: number of turns to wait after start_func before re-enabling input
        :param end_func: optional. Function to call after waiting ``turns`` after start_func
        :return: ``None``

        """
		self.disable_input()
		self.app.wait_travel_command(character, thing, dest, start_func, turns,
										partial(self.enable_input, end_func))


class Screens(Widget):
	"""Children of this widget will be added to the ``GameApp``'s :class:`kivy.uix.screenmanager.ScreenManager`"""
	app = ObjectProperty()

	def add_widget(self, wid, index=0, canvas=None):
		wid.switch_screen = self.app.screen_manager.setter('screen')
		super().add_widget(wid, index, canvas)


class GameApp(ELiDEApp):
	modules = []
	world_file = None
	turn_length = NumericProperty(0.5)

	def wait_turns(self, turns, dt=None, *, cb=None):
		"""Call ``self.engine.next_turn()`` ``n`` times, waiting ``self.turn_length`` in between

        If provided, call ``cb`` when done.

        :param turns: number of turns to wait
        :param dt: unused, just satisfies the clock
        :param cb: callback function to call when done, optional
        :return: ``None``

        """
		if turns == 0:
			if cb:
				cb()
			return
		self.engine.next_turn()
		turns -= 1
		Clock.schedule_once(partial(self.wait_turns, turns, cb=cb),
							self.turn_length)

	def wait_travel(self, character, thing, dest, cb=None):
		"""Schedule a thing to travel someplace, then wait for it to finish, and call ``cb`` if provided

        :param character: name of the character
        :param thing: name of the thing
        :param dest: name of the destination (a place)
        :param cb: function to be called when I'm done
        :return: ``None``
        
        """
		self.wait_turns(
			self.engine.character[character].thing[thing].travel_to(dest),
			cb=cb)

	def wait_command(self, start_func, turns=1, end_func=None):
		"""Call ``start_func``, and wait to call ``end_func`` after simulating ``turns`` (default 1)

        :param start_func: function to call before waiting
        :param turns: number of turns to wait
        :param end_func: function to call after waiting
        :return: ``None``

        """
		start_func()
		self.wait_turns(turns, cb=end_func)

	def wait_travel_command(self,
							character,
							thing,
							dest,
							start_func,
							turns=1,
							end_func=None):
		"""Schedule a thing to travel someplace and do something, then wait for it to finish.

        :param character: name of the character
        :param thing: name of the thing
        :param dest: name of the destination (a place)
        :param start_func: function to call when the thing gets to dest
        :param turns: number of turns to wait after start_func before re-enabling input
        :param end_func: optional. Function to call after waiting ``turns`` after start_func
        :return: ``None``
        """
		self.wait_travel(character,
							thing,
							dest,
							cb=partial(self.wait_command, start_func, turns,
										end_func))

	def on_engine(self, *args):
		self.branch, self.turn, self.tick = self.engine._btt()
		self.engine.time.connect(self._pull_time, weak=False)

	def _pull_time(self, *args, branch, turn, tick):
		self.branch, self.turn, self.tick = branch, turn, tick

	def _get_worlddb(self):
		filen = self.world_file or \
                      self.name + 'World.db' if self.name \
               else 'LiSEWorld.db'
		return resource_find(filen) or filen

	worlddb = AliasProperty(_get_worlddb, lambda self, v: None)

	def build(self):
		have_world = False
		try:
			os.stat(self.worlddb)
			have_world = True
		except FileNotFoundError:
			pass
		self.procman = LiSE.proxy.EngineProcessManager()
		self.engine = self.procman.start(self.worlddb,
											logger=Logger,
											loglevel=getattr(
												self, 'loglevel', 'debug'),
											do_game_start=not have_world,
											install_modules=self.modules)
		self.screen_manager = ScreenManager()
		self.screens = Screens(app=self)
		self.screens.bind(children=self._pull_screens)
		self._pull_screens()
		if hasattr(self, 'inspector'):
			from kivy.core.window import Window
			from kivy.modules import inspector
			inspector.create_inspector(Window, self.screen_manager)
		return self.screen_manager

	@trigger
	def _pull_screens(self, *args):
		for screen in reversed(self.screens.children):
			self.screens.remove_widget(screen)
			self.screen_manager.add_widget(screen)

	def on_pause(self):
		"""Sync the database with the current state of the game."""
		self.engine.commit()
		self.config.write()

	def on_stop(self, *largs):
		"""Sync the database, wrap up the game, and halt."""
		self.procman.shutdown()
		self.config.write()



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial
from kivy.properties import (DictProperty, ObjectProperty, OptionProperty,
								ListProperty, BooleanProperty,
								BoundedNumericProperty, NumericProperty,
								StringProperty)
from kivy.uix.widget import Widget
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.label import Label
from kivy.uix.slider import Slider
from kivy.uix.textinput import TextInput
from kivy.uix.modalview import ModalView
from kivy.uix.gridlayout import GridLayout
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview import RecycleView
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.clock import Clock
from kivy.lang import Builder

from ELiDE.util import trigger


class CalendarWidget(RecycleDataViewBehavior, Widget):
	"""Base class for widgets within a Calendar

    Shows the value of its ``key`` at a particular ``turn``, and sets
    it at that turn if the value changes.

    """
	turn = NumericProperty()
	"""What turn I'm displaying the stat's value for"""
	key = ObjectProperty()
	"""The key to set in the entity"""
	value = ObjectProperty(allownone=True)
	"""The value you want to set the key to"""

	def _update_disabledness(self, *args, **kwargs):
		if not self.parent:
			return
		self.disabled = self.turn < self.parent.parent.entity.engine.turn

	def _trigger_update_disabledness(self, *args, **kwargs):
		if hasattr(self, '_scheduled_update_disabledness'):
			Clock.unschedule(self._scheduled_update_disabledness)
		self._scheduled_update_disabledness = Clock.schedule_once(
			self._update_disabledness)

	def _set_value(self):
		entity = self.parent.parent.entity
		entity = getattr(entity, 'stat', entity)
		entity[self.key] = self.value

	def on_value(self, *args):
		# do I want to do some validation at this point?
		# Maybe I should validate on the proxy objects and catch that in Calendar,
		# display an error message?
		if not self.parent:
			return
		calendar = self.parent.parent
		my_dict = calendar.idx[(self.turn, self.key)]
		entity = calendar.entity
		update_mode = calendar.update_mode
		if my_dict['value'] != self.value:
			my_dict['value'] = self.value
			if update_mode == 'batch':
				calendar.changed = True
			elif update_mode == 'present':
				if self.turn == entity.engine.turn:
					self._set_value()
				else:
					calendar.changed = True
			else:
				eng = entity.engine
				now = eng.turn
				if now == self.turn:
					self._set_value()
				else:
					eng.turn = self.turn
					self._set_value()
					eng.turn = now

	def on_parent(self, *args):
		if not self.parent:
			return
		self._trigger_update_disabledness()
		self.parent.parent.entity.engine.time.connect(
			self._trigger_update_disabledness)


class CalendarLabel(CalendarWidget, Label):

	def __init__(self, **kwargs):
		if 'text' not in kwargs or not kwargs['text']:
			kwargs['text'] = ''
		super().__init__(**kwargs)


class CalendarSlider(Slider, CalendarWidget):
	pass


class CalendarTextInput(CalendarWidget, TextInput):

	def _parse_text(self, *args):
		from ast import literal_eval
		try:
			v = literal_eval(self.text)
		except (TypeError, ValueError, SyntaxError):
			v = self.text
		self.value = self.hint_text = v
		self.text = ''

	_trigger_parse_text = trigger(_parse_text)


class CalendarOptionButton(CalendarWidget, Button):
	options = ListProperty()
	modalview = ObjectProperty()
	cols = BoundedNumericProperty(1, min=1)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._make_modalview()
		self._update_modalview()
		self.bind(cols=self._make_modalview)
		self.bind(options=self._update_modalview)
		self.bind(on_release=self.modalview.open)

	def _make_modalview(self, *args):
		if not self.modalview:
			self.modalview = ModalView()
		if self.modalview.children:
			container = self.modalview.children[0]
		else:
			container = GridLayout(cols=self.cols)
			self.modalview.add_widget(container)
		container.size = container.minimum_size
		self._update_modalview()

	def _update_modalview(self, *args):
		if not self.modalview:
			Clock.schedule_once(self.on_options, 0)
			return
		if not self.modalview.children:
			container = GridLayout(cols=self.cols)
			self.modalview.add_widget(container)
		else:
			container = self.modalview.children[0]
		for option in self.options:
			if type(option) is tuple:
				text, value = option
				container.add_widget(
					Button(size_hint_y=None,
							height=30,
							text=text,
							on_release=partial(self._set_value_and_close,
												text)))
			else:
				container.add_widget(
					Button(text=str(option),
							on_release=partial(self._set_value_and_close,
												str(option)),
							size_hint_y=None,
							height=30))
		container.size = container.minimum_size

	def _set_value_and_close(self, val, *args):
		self.value = val
		self.modalview.dismiss()


class CalendarToggleButton(CalendarWidget, ToggleButton):
	index = None
	true_text = StringProperty('True')
	false_text = StringProperty('False')

	def on_state(self, *args):
		self.value = self.state == 'down'
		self.text = self.true_text if self.value else self.false_text


class CalendarMenuLayout(LayoutSelectionBehavior, RecycleBoxLayout):
	pass


class AbstractCalendar(RecycleView):
	_control2wid = {
		'slider': 'CalendarSlider',
		'togglebutton': 'CalendarToggleButton',
		'textinput': 'CalendarTextInput',
		'option': 'CalendarOptionButton'
	}
	cols = NumericProperty(1)
	"""Number of columns to display, default 1"""
	entity = ObjectProperty()
	"""The LiSE proxy object to display the stats of"""
	idx = DictProperty()
	"""Dictionary mapping ``key, turn`` pairs to their widgets"""
	changed = BooleanProperty(False)
	"""Whether there are changes yet to be committed to the LiSE core"""
	update_mode = OptionProperty('batch', options=['batch', 'present', 'all'])
	"""How to go about submitting changes to the LiSE core. Options:
    
    * ``'batch'`` (default): don't submit changes automatically. You have to call
    ``get_track`` and apply the changes using the ``LiSE.handle`` method
    ``apply_choices``, eg.
        ``engine_proxy.handle('apply_choices', choices=calendar.get_track())``
    * ``'present'``: immediately apply changes that affect the current turn,
    possibly wiping out anything in the future -- so you still have to use
    ``get_track`` and ``apply_choices``. However, if you're using a calendar
    in the same interface as another control widget for the same stat,
    ``'present'`` will ensure that the two widgets always display the same value.
    * ``'all'``: apply every change immediately to the LiSE core. Should only be
    used when the LiSE core is in planning mode.
    
    """
	headers = BooleanProperty(True)
	"""Whether to display the name of the stat above its column, default ``True``"""
	turn_labels = BooleanProperty(True)
	"""Whether to display the turn of the value before its row, default ``True``"""
	turn_label_transformer = ObjectProperty(str)
	"""A function taking the turn number and returning a string to represent it
    
    Defaults to ``str``, but you might use this to display eg. the day of the
    week instead.
    
    """

	def on_data(self, *args):
		idx = self.idx
		for item in self.data:
			if 'key' in item and 'turn' in item:
				idx[(item['turn'], item['key'])] = item

	def get_track(self):
		"""Get a dictionary that can be used to submit my changes to ``LiSE.Engine.apply_choices``

        If a data dictionary does not have the key 'turn', it will not be included in the track.
        You can use this to add labels and other non-input widgets to the calendar.

        """
		changes = []
		track = {'entity': self.entity, 'changes': changes}
		if not self.data:
			return track
		for datum in self.data:
			if 'turn' in datum:
				break
		else:
			# I don't know *why* the calendar has no actionable data in it but here u go
			return track
		last = self.entity.engine.turn
		accumulator = []
		for datum in self.data:
			if 'turn' not in datum:
				continue
			trn = datum['turn']
			if trn < last:
				continue
			if trn > last:
				if trn > last + 1:
					changes.extend([[]] * (trn - last - 1))
				changes.append(accumulator)
				accumulator = []
				last = trn
			accumulator.append((datum['key'], datum['value']))
		changes.append(accumulator)
		return track


class Agenda(AbstractCalendar):

	def from_schedule(self, schedule, start_turn=None, key=str):
		# It should be convenient to style the calendar using data from the core;
		# not sure what the API should be like
		control2wid = self._control2wid
		if start_turn is None:
			start_turn = self.entity.engine.turn
		curturn = start_turn
		endturn = curturn + len(next(iter(schedule.values())))
		data = []
		stats = sorted((stat for stat in schedule if not stat.startswith('_')),
						key=key)
		headers = self.headers
		turn_labels = self.turn_labels
		if headers:
			if turn_labels:
				data.append({'widget': 'Label', 'text': ''})
			for stat in stats:
				if stat.startswith('_'):
					continue
				data.append({
					'widget': 'Label',
					'text': str(stat),
					'bold': True
				})
		cols = len(data)
		iters = {stat: iter(values) for (stat, values) in schedule.items()}
		for turn in range(curturn, endturn):
			if turn_labels:
				data.append({
					'widget': 'Label',
					'text': self.turn_label_transformer(turn),
					'bold': True
				})
			if '_config' in iters:
				config = next(iters['_config'])
			else:
				config = None
			for stat in stats:
				datum = {'key': stat, 'value': next(iters[stat]), 'turn': turn}
				if config and stat in config and 'control' in config[stat]:
					datum.update(config[stat])
					datum['widget'] = control2wid.get(
						datum.pop('control', None), 'CalendarLabel')
					if datum['widget'] == 'CalendarToggleButton':
						if datum['value']:
							datum['text'] = config[stat].get(
								'true_text', 'True')
							datum['state'] = 'down'
						else:
							datum['text'] = config[stat].get(
								'false_text', 'False')
							datum['state'] = 'normal'
					elif datum['widget'] == 'CalendarTextInput':
						datum['hint_text'] = str(datum['value'])
				else:
					datum['widget'] = 'CalendarLabel'
				data.append(datum)
		(self.cols, self.data, self.changed) = (cols, data, False)


class Calendar(AbstractCalendar):
	multicol = BooleanProperty(False)

	def from_schedule(self, schedule, start_turn=None, key=str):
		control2wid = self._control2wid
		if start_turn is None:
			start_turn = self.entity.engine.turn
		curturn = start_turn
		endturn = curturn + len(next(iter(schedule.values())))
		data = []
		stats = sorted((stat for stat in schedule if not stat.startswith('_')),
						key=key)
		iters = {stat: iter(values) for (stat, values) in schedule.items()}
		headers = self.headers
		turn_labels = self.turn_labels
		for turn in range(curturn, endturn):
			if turn_labels:
				data.append({
					'widget': 'Label',
					'text': self.turn_label_transformer(turn),
					'bold': True
				})
			if '_config' in iters:
				config = next(iters['_config'])
			else:
				config = None
			for stat in stats:
				if headers:
					data.append({
						'widget': 'Label',
						'text': str(stat),
						'bold': True
					})
				datum = {'key': stat, 'value': next(iters[stat]), 'turn': turn}
				if config and stat in config and 'control' in config[stat]:
					datum.update(config[stat])
					datum['widget'] = control2wid.get(
						datum.pop('control', None), 'CalendarLabel')
					if datum['widget'] == 'CalendarToggleButton':
						if datum['value']:
							datum['text'] = config[stat].get(
								'true_text', 'True')
							datum['state'] = 'down'
						else:
							datum['text'] = config[stat].get(
								'false_text', 'False')
							datum['state'] = 'normal'
					elif datum['widget'] == 'CalendarTextInput':
						datum['hint_text'] = str(datum['value'])
				else:
					datum['widget'] = 'CalendarLabel'
				data.append(datum)
		if self.multicol:
			self.cols = endturn - curturn
		self.data = data


Builder.load_string("""
<Agenda>:
    key_viewclass: 'widget'
    RecycleGridLayout:
        cols: root.cols
        size_hint: None, None
        default_size: dp(84), dp(36)
        default_size_hint: None, None
        size: self.minimum_size
        orientation: 'tb-lr'
<Calendar>:
    turn_labels: False
    key_viewclass: 'widget'
    RecycleGridLayout:
        cols: 1
        orientation: 'lr-tb'
        default_size_hint: 1, None
        default_size: dp(84), dp(36)
        size: self.minimum_size
<CalendarLabel>:
    text: str(self.value) if self.value is not None else ''
<CalendarSlider>:
    padding: 5
    Label:
        x: root.center_x - (self.width / 2)
        y: root.center_y
        text: str(root.value)
        size: self.texture_size
<CalendarOptionButton>:
    text: str(self.value)
<CalendarTextInput>:
    multiline: False
    on_text_validate: self._trigger_parse_text()
""")

if __name__ == '__main__':
	from kivy.app import App

	class CalendarTestApp(App):

		def build(self):
			self.wid = Calendar()
			return self.wid

		def on_start(self):
			# it seems like the calendar blanks out its data sometime after initialization
			data = []
			for i in range(7):
				for j in range(3):
					data.append({'widget': 'Button', 'text': f'row{i} col{j}'})
			self.wid.data = data

	CalendarTestApp().run()



from kivy.tests.common import UnitTestTouch

from abc import abstractmethod
from LiSE import Engine
from LiSE.examples import kobold
from LiSE.examples import polygons
from .util import idle_until, window_with_widget, ELiDEAppTest
from ..card import Card


class RuleBuilderTest(ELiDEAppTest):

	@abstractmethod
	def install(self, engine: Engine):
		raise NotImplementedError()

	@abstractmethod
	def get_selection(self):
		raise NotImplementedError()

	def setUp(self):
		super(RuleBuilderTest, self).setUp()
		with Engine(self.prefix) as eng:
			self.install(eng)
		app = self.app
		mgr = app.build()
		self.win = window_with_widget(mgr)
		screen = app.mainscreen
		idle_until(lambda: app.rules.rulesview, 100, 'Never made rules view')
		idle_until(lambda: 'physical' in screen.graphboards, 100,
					'Never made physical board')
		self.board = screen.graphboards['physical']
		idle_until(lambda: 'kobold' in self.board.pawn, 100,
					'never got the pawn for the kobold')
		app.selection = self.get_selection()
		screen.charmenu.charmenu.toggle_rules()
		rules = app.rules
		rules_box = rules.children[0]
		idle_until(lambda: 'ruleslist' in rules_box.ids, 100,
					'Never made rules list')
		self.rules_list = rules_list = rules_box.ids.ruleslist
		self.rules_view = rules_box.ids.rulesview
		idle_until(lambda: rules_list.children[0].children, 100,
					'Never filled rules list')


class TestRuleBuilderKobold(RuleBuilderTest):

	def install(self, engine: Engine):
		kobold.inittest(engine)

	def get_selection(self):
		return self.board.pawn['kobold']

	def test_rule_builder_display_trigger(self):
		rules_list = self.rules_list
		rules_view = self.rules_view
		idle_until(
			lambda: 'shrubsprint' in
			{rulebut.text
				for rulebut in rules_list.children[0].children}, 100,
			'Never made shrubsprint button')
		for rulebut in rules_list.children[0].children:
			if rulebut.text == 'shrubsprint':
				rulebut.state = 'down'
				break
		idle_until(lambda: rules_view.children)
		idle_until(lambda: hasattr(rules_view, '_trigger_tab'), 100,
					'Never made trigger tab')
		builder = rules_view._trigger_builder
		idle_until(
			lambda:
			[child for child in builder.children if isinstance(child, Card)],
			100, 'Never filled trigger builder')
		card_names = {
			card.headline_text
			for card in builder.children if isinstance(card, Card)
		}
		assert card_names == {
			'standing_still', 'aware', 'uncovered', 'sametile', 'breakcover',
			'kobold_alive'
		}

	def test_rule_builder_remove_trigger(self):
		rules_list = self.rules_list
		rules_view = self.rules_view
		idle_until(
			lambda: 'shrubsprint' in
			{rulebut.text
				for rulebut in rules_list.children[0].children}, 100,
			'Never made shrubsprint button')
		for rulebut in rules_list.children[0].children:
			if rulebut.text == 'shrubsprint':
				rulebut.state = 'down'
				break
		idle_until(lambda: rules_view.children)
		idle_until(lambda: hasattr(rules_view, '_trigger_tab'), 100,
					'Never made trigger tab')
		builder = rules_view._trigger_builder
		idle_until(
			lambda:
			[child for child in builder.children if isinstance(child, Card)],
			100, 'Never filled trigger builder')
		for card in builder.children:
			if not isinstance(card, Card):
				continue
			if card.headline_text == 'breakcover':
				break
		else:
			assert False, "No breakcover card"
		for foundation in builder.children:
			if isinstance(foundation, Card):
				continue
			if foundation.x > card.right:
				break
		else:
			assert False, "No right foundation"
		mov = UnitTestTouch(*card.center)
		mov.touch_down()
		dist_x = foundation.center_x - card.center_x
		dist_y = foundation.y - card.center_y
		for i in range(1, 11):
			coef = 1 / i
			x = foundation.center_x - coef * dist_x
			y = foundation.y - coef * dist_y
			mov.touch_move(x, y)
			self.advance_frames(1)
		mov.touch_up(foundation.center_x, foundation.y)
		idle_until(lambda: card.x == foundation.x, 100, "card didn't move")
		idle_until(
			lambda: 'breakcover' not in self.app.engine.rule['shrubsprint'].
			triggers, 100, 'breakcover never removed from rulebook')

	def test_rule_builder_add_trigger(self):
		rules_list = self.rules_list
		rules_view = self.rules_view
		idle_until(
			lambda: 'shrubsprint' in
			{rulebut.text
				for rulebut in rules_list.children[0].children}, 100,
			'Never made shrubsprint button')
		for rulebut in rules_list.children[0].children:
			if rulebut.text == 'shrubsprint':
				rulebut.state = 'down'
				break
		idle_until(lambda: rules_view.children)
		idle_until(lambda: hasattr(rules_view, '_trigger_tab'), 100,
					'Never made trigger tab')
		builder = rules_view._trigger_builder
		idle_until(
			lambda:
			[child for child in builder.children if isinstance(child, Card)],
			100, 'Never filled trigger builder')
		aware = breakcover = None
		for card in builder.children:
			if not isinstance(card, Card):
				continue
			if card.headline_text == 'aware':
				aware = card
			elif card.headline_text == 'breakcover':
				breakcover = card
		assert None not in (
			aware, breakcover), "Didn't get 'aware' and 'breakcover' cards"
		start_x = aware.center_x
		start_y = aware.top - 10
		mov = UnitTestTouch(start_x, start_y)
		mov.touch_down()
		dist_x = start_x - breakcover.center_x
		dist_y = start_y - breakcover.center_y
		decr_x = dist_x / 10
		decr_y = dist_y / 10
		x = start_x
		y = start_y
		for i in range(1, 11):
			x -= decr_x
			y -= decr_y
			mov.touch_move(x, y)
			self.advance_frames(1)
		mov.touch_up(*breakcover.center)
		idle_until(lambda: aware.x == breakcover.x, 100,
					"aware didn't move to its new place")
		idle_until(
			lambda: 'aware' in self.app.engine.rule['shrubsprint'].triggers,
			100, 'aware never added to rulebook')


class TestCharRuleBuilder(ELiDEAppTest):

	def setUp(self):
		super(TestCharRuleBuilder, self).setUp()
		with Engine(self.prefix) as eng:
			polygons.install(eng)
			assert list(
				eng.character['triangle'].unit.rule['relocate'].triggers) == [
					eng.trigger.similar_neighbors,
					eng.trigger.dissimilar_neighbors
				]
		app = self.app
		mgr = app.build()
		self.win = window_with_widget(mgr)
		idle_until(lambda: getattr(app, 'engine'), 100,
					'App never made engine')
		idle_until(lambda: 'triangle' in app.engine.character, 100,
					'Engine proxy never made triangle character proxy')
		app.select_character(app.engine.character['triangle'])
		idle_until(lambda: app.character_name == 'triangle', 100,
					'Never changed character')
		app.mainscreen.charmenu.charmenu.toggle_rules()
		idle_until(lambda: getattr(app.charrules, '_finalized', False), 100,
					'Never finalized')

	def test_char_rule_builder_remove_unit_trigger(self):
		app = self.app
		idle_until(lambda: getattr(app.charrules, '_finalized', False), 100,
					"Never finalized charrules")
		tabitem = app.charrules._unit_tab
		idle_until(lambda: tabitem.content, 100, 'unit tab never got content')
		tabitem.on_press()
		self.advance_frames(1)
		tabitem.on_release()
		idle_until(lambda: app.charrules._tabs.current_tab == tabitem, 100,
					'Never switched tab')
		rules_box = app.charrules._unit_box
		idle_until(lambda: rules_box.parent, 100, 'unit box never got parent')
		idle_until(lambda: getattr(rules_box.rulesview, '_finalized', False),
					100, "Never finalized unit rules view")
		idle_until(lambda: rules_box.children, 100,
					'_unit_box never got children')
		idle_until(lambda: rules_box.rulesview.children, 100,
					'Never filled rules view')
		rules_list = rules_box.ruleslist
		idle_until(lambda: rules_list.children[0].children, 100,
					'Never filled rules list')
		idle_until(
			lambda: 'relocate' in
			{rulebut.text
				for rulebut in rules_list.children[0].children}, 100,
			'Never made relocate button')
		for rulebut in rules_list.children[0].children:
			if rulebut.text == 'relocate':
				rulebut.state = 'down'
				break
		builder = rules_box.rulesview._trigger_builder
		assert rules_box.rulesview._tabs.current_tab == rules_box.rulesview._trigger_tab
		idle_until(lambda: builder.children, 100,
					'trigger builder never got children')

		def builder_foundation():
			for child in builder.children:
				if not isinstance(child, Card):
					return True
			return False

		idle_until(builder_foundation, 100, 'Never filled trigger builder')
		idle_until(lambda: builder.parent, 100,
					"trigger builder never got parent")
		card_names = {
			card.headline_text
			for card in builder.children if isinstance(card, Card)
		}
		assert card_names == {'similar_neighbors', 'dissimilar_neighbors'}
		for card in builder.children:
			if not isinstance(card, Card):
				continue
			if card.headline_text == 'similar_neighbors':
				break
		else:
			assert False, "Didn't get similar_neighbors"
		startx = card.center_x
		starty = card.top - 1
		assert card.collide_point(startx, starty), "card didn't collide itself"
		for cardother in builder.children:
			if not isinstance(cardother, Card) or cardother == card:
				continue
			assert not cardother.collide_point(
				startx, starty), "other card will grab the touch"
		touch = UnitTestTouch(startx, starty)
		for target in builder.children:
			if isinstance(target, Card):
				continue
			if target.x > card.right:
				break
		else:
			assert False, "Didn't get target foundation"
		targx, targy = target.center
		distx = targx - startx
		disty = targy - starty
		x, y = startx, starty
		touch.touch_down()
		for i in range(1, 11):
			x += distx / 10
			y += disty / 10
			touch.touch_move(x, y)
			self.advance_frames(1)
		touch.touch_up()
		self.advance_frames(5)
		rules_box.ids.closebut.on_release()
		self.advance_frames(5)
		app.stop()
		with Engine(self.prefix) as eng:
			assert list(
				eng.character['triangle'].unit.rule['relocate'].triggers) == [
					eng.trigger.dissimilar_neighbors
				]
